-- VECTRIC LUA SCRIPT
-- Copyright 2013 Vectric Ltd.

-- Gadgets are an entirely optional add-in to Vectric's core software products. 
-- They are provided 'as-is', without any express or implied warranty, and you make use of them entirely at your own risk.
-- In no event will the author(s) or Vectric Ltd. be held liable for any damages arising from their use.

-- Permission is granted to anyone to use this software for any purpose, 
-- including commercial applications, and to alter it and redistribute it freely, 
-- subject to the following restrictions:

-- 1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
-- 2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
-- 3. This notice may not be removed or altered from any source distribution.

--[[  -------------- Shopbot Probe Importer --------------------------------------------------
|
|  A Lua script to import shopbot probe toolpaths into Aspire
]]

require "strict"
--require("mobdebug").start()

g_title = "Shopbot Probe Importer"
g_version = "1.2"

--[[  -------------- Trim --------------------------------------------------
|
|  Trims whitespace from both ends of a string
|  http://snippets.luacode.org/snippets/trim_whitespace_from_string_76
|  Parameters:
|	  s          -- The string to trim
|     
|  Return Values:
|  A trimmed version of the passed string
]]
function Trim(s)
  return s:find('^%s*$') and '' or s:match('^%s*(.*%S)')
end

--[[  -------------- StartsWith --------------------------------------------------
|
|  Checks to see if a string starts with a prefix
|  http://lua-users.org/wiki/StringRecipes
|  Parameters:
|	  s        -- The string to test
|   prefix     -- The prefix to find
|     
|  Return Values:
|  True if the prefix matches
]]
function StartsWith(s, prefix)
   return string.sub(s, 1, string.len(prefix)) == prefix
end


--[[  -------------- ParseLine --------------------------------------------------
|
|  Parameters:
|	  line      -- The line to parse
|   variables -- Table of variables we know about
|     
|  Return Values:
|  The name of the command and a table of arguments
]]
function ParseLine(line, variables)
-- Assume line is trimmed
-- Find first comma, if there isn't one then the command doesn't take any arguments
local comma_index = line:find(",", 1, true)
if not comma_index then
	return line, {}
end
-- We have a comma so build up a list of arguments
local command = line:sub(1, comma_index - 1)
command = Trim(command)
local arguments = {}
-- Keep finding arguments
local done = false
while not done do
	-- Find the next delimiter
  -- If we can't find one then we need to the end of the string
	local next_comma_index =  line:find(",", comma_index + 1, true)
  if not next_comma_index then
    next_comma_index = line:len() + 1
    done = true
  end
	-- Extract the argument
	local argument = line:sub(comma_index + 1, next_comma_index - 1)
	argument = Trim(argument)
  local argv = StartsWith(argument, '&') and variables[argument] or tonumber(argument)
  -- If we can't get the value then that's an error
  if not argv then
    return nil, {}
  end
  -- Assign the value
  arguments[#arguments + 1] = argv
	-- Move the comma position along
	comma_index = next_comma_index
end
return command, arguments
end

--[[  -------------- ParseVariableAssignment --------------------------------------------------
|
|  Parameters:
|	  line     -- The line to parse for variable assignment
|     
|  Return Values:
|  The name of the variable and its value
]]
function ParseVariableAssignment(line)
-- Assume that the line has both been trimmed
-- and starts with &
local equals_index = line:find("=", 1, true)
if not equals_index then
  return nil, nil
end
local name = Trim(line:sub(1, equals_index - 1))
local value = Trim(line:sub(equals_index + 1))
return name, value and tonumber(value) or nil
end


--[[  -------------- UnsupportedCommand --------------------------------------------------
|
|  Parameters:
|	command         -- The command we don't support
|   line_number     -- The line number it occured on
|     
]]
function UnsupportedCommand(command, line_number)
local format_spec = "Unsupported command: %s, Line %d."
local message = string.format(format_spec, command, line_number)
MessageBox(message)
end

--[[  -------------- UnknownCommand --------------------------------------------------
|
|  Parameters:
|	command         -- The command we don't know about but might want to suppot
|   line_number     -- The line number it occured on
|     
]]
function UnknownCommand(command, line_number)
local format_spec = "Unknown command: %s, Line %d."
local message = string.format(format_spec, command, line_number)
MessageBox(message)
end

--[[  -------------- SyntaxError --------------------------------------------------
|
|  Parameters:
|	 line          -- The command we don't know about but might want to suppot
|  line_number   -- The line number it occured on
|     
]]
function SyntaxError(line, line_number)
local format_spec = "Syntax Error: %s, Line %d."
local message = string.format(format_spec, line, line_number)
MessageBox(message)
end


--[[  -------------- CreateContourFromSBPFile --------------------------------------------------
|
|  Parses an SBP file returning a contour.
|  The SBP parser is very basic and supports a limited subset of commands.
|
|  Ignores:
|  PAUSE
|  SA, SO
|
|  Supports:
|  J2, J3, JZ
|  M2, M3, MZ
|  Variable assignment
|
|  Unsupported:
|  SR
|
|
|  Parameters:
|  	filename   -- The filename of the SBP command
|     
|  Return Values:
|  A contour representing the toolpath, nil in the case of failure
]]
function CreateContourFromSBPFile(filename)
--unpack update for lua 5.2
local unpack = table.unpack  
-- Open the file
local file = io.open(filename, "r")
if not file then
	return nil
end

-- Table of variables
local variables = {}

-- Create a contour to hold the toolpath
local contour = Contour(0.0)
contour:AppendPoint(0, 0)

-- Read each line in the file
local line_number = 1
while true do
	-- Read and check for line and trim away any whitespace
	local line = file:read("*line")
	if not line then
		break
	end
	local trimmed_line = Trim(line)
	--Deal with each of the cases
	if trimmed_line:len() == 0 then
		-- Empty line
	elseif StartsWith(trimmed_line, '\'') then
		-- Comment
	elseif StartsWith(trimmed_line, "PAUSE") then
		-- Pause doesn't fit the 2 letter mnemonic comma separated list scheme
  elseif StartsWith(trimmed_line, '&') then
    -- Variable assignment
    local name, value = ParseVariableAssignment(trimmed_line)
    if name and value then
      variables[name] = value
    else
      SyntaxError(trimmed_line, line_number)
      return nil
    end
	else
		-- Possible command
		local command, arguments = ParseLine(trimmed_line, variables)
    if not command then
      -- Nil command signals syntax error
      SyntaxError(trimmed_line, line_number)
      return nil
		elseif command == "SA" then
			-- Set absolute distances
      -- Assume this is the case
		elseif command == "SO" then
			-- Set output switch
    elseif command == "SR" then
      -- Set relative distances
      -- Unsupported
      UnsupportedCommand(command, line_number)
      return nil
    elseif command == "J2" then
			-- Jog 2 dimensions
      local point = contour.EndPoint3D
      local x, y = unpack(arguments)
      contour:LineTo(x, y, point.z)
    elseif command == "J3" then
      -- Jog 3 dimensions
      local x, y, z = unpack(arguments)
      contour:LineTo(x, y, z)
		elseif command == "JZ" then
			-- Jog Z
      local point = contour.EndPoint3D
      local z = unpack(arguments)
      contour:LineTo(point.x, point.y, z)
     elseif command == "M2" then
			-- Move 2 dimensions
      local point = contour.EndPoint3D
      local x, y = unpack(arguments)
      contour:LineTo(x, y, point.z)
    elseif command == "M3" then
			-- Move 3 dimensions
      local x, y, z = unpack(arguments)
      contour:LineTo(x, y, z)
    elseif command == "MZ" then
      -- Move Z
      local point = contour.EndPoint3D
      local z = unpack(arguments)
      contour:LineTo(point.x, point.y, z)
		else
			-- Unknown command
      UnknownCommand(command, line_number)
      return nil
		end
	end
	-- Bump the line number
	line_number = 1 + line_number
end

-- Close the file
file:close()

-- Return the contour
return contour
end

--[[  -------------- CreateToolpath --------------------------------------------------
|
|  Create an 'imported' toolpath within the program.
|  
|  Parameters:
|	  options           -- The set of options for the toolpath
|   vectors_group     -- The vectors representing the toolpath
|     
|
|  Return Values:
|     true if toolpath created OK else false
|
]]
function CreateToolpath(options, vector_group)

-- Unpack the relevant options
local toolpath_name = options.toolpath_name
local tool          = options.tool
local start_depth   = options.start_depth

-- Create object used to set home position and safe z gap above material surface
local pos_data = ToolpathPosData()
  
-- object used to pass data on toolpath settings
local toolpath_options = ExternalToolpathOptions()
toolpath_options.StartDepth = start_depth
toolpath_options.CreatePreview = true

-- Create our toolpath
local toolpath = ExternalToolpath(
                                  toolpath_name,
                                  tool,
                                  pos_data,
                                  toolpath_options,
                                  vector_group
                                  )

if toolpath:Error() then
  DisplayMessageBox("Error creating toolpath")
  return false
end

local toolpath_manager = ToolpathManager()

return toolpath_manager:AddExternalToolpath(toolpath)

end

--[[  -------------- DisplayDialog --------------------------------------------------
|
|  Displays the options dialog
|  Parameters:
|	options           -- The set of options for dialog
|   script_path       -- Where to locate the html pages
|   job            	  -- The job
|     
|
|  Return Values:
|  1 on success,
|  0 if the user cancelled
|  -1 on bad input
|
]]
function DisplayDialog(options, script_path, job)

  -- Prepare the dialog
  local script_html = "file:" .. script_path .. "\\Shopbot_Probe_Importer.htm"
  local dialog_name = options.gadget_title
  local dialog = HTML_Dialog(false, script_html, options.window_width, options.window_height, dialog_name)
  
  -- Set the gadget name and version
  dialog:AddLabelField("GadgetTitle", options.gadget_title)
  dialog:AddLabelField("GadgetVersion", options.gadget_version)
  
  -- Add a toolpath name field
  dialog:AddTextField("ToolpathNameEdit", options.toolpath_name)
  
  -- Add the tool picker
  dialog:AddLabelField("ToolNameLabel", "")
  dialog:AddToolPicker("ChooseToolButton", "ToolNameLabel", options.tool_id)
  dialog:AddToolPickerValidToolType("ChooseToolButton", Tool.BALL_NOSE)
  dialog:AddToolPickerValidToolType("ChooseToolButton", Tool.END_MILL)
  dialog:AddToolPickerValidToolType("ChooseToolButton", Tool.RADIUSED_END_MILL)
  
  -- Add file picker
  dialog:AddTextField("FileNameEdit", options.filename)
  dialog:AddFilePicker(true, "ChooseFileButton", "FileNameEdit", true)
  
  -- Add z zero set to table checkbox
  dialog:AddCheckBox("ZZeroSetToTableCheck", options.z_zero_set_to_table)
  
  -- Show the dialog
  if not dialog:ShowDialog() then
    return 0
  end
	
  -- Update the options from this dialog
  options.toolpath_name = dialog:GetTextField("ToolpathNameEdit")
  options.tool = dialog:GetTool("ChooseToolButton")
  options.tool_id = options.tool:ToolDBId
  options.filename = dialog:GetTextField("FileNameEdit")
  options.z_zero_set_to_table = dialog:GetCheckBox("ZZeroSetToTableCheck")
  
  -- Validate the string input
  local StringInputInvalid = function(input)
    return not input or input:len() == 0
  end
  
  if StringInputInvalid(options.toolpath_name) then
    return -1
  end

  if StringInputInvalid(options.tool.Name) then
    return -1
  end
  
  if StringInputInvalid(options.filename) then
    return -1
  end
  
  if not options.tool then
	return -1
  end
  
  
  return 1
end


--[[  ------------------------ main --------------------------------------------------
|
|  Entry point for script
|
]]
function main(script_path)

-- Check that we're aspire
if not IsAspire() then
	MessageBox("This gadget is Aspire only!")
	return false
end

-- Check for the existence of a job
local job = VectricJob()
if not job.Exists then
  	MessageBox("No job created, please start a new job and re-run the gadget")
  return false
end
  
-- Set up our options and display the dialog
local options = {}
options.gadget_title = g_title
options.gadget_version = g_version
options.window_width = 660
options.window_height = 280
options.filename = ""
options.z_zero_set_to_table = false
  
-- Toolpath and tool settings
options.toolpath_name = "Shopbot Probe Toolpath"
options.tool_id = ToolDBId()
options.tool = nil

-- Start depth options
options.start_depth = 0.0
  
-- Loop until either the user cancels or presses ok
local dialog_code = -1
while dialog_code == -1 do
	dialog_code = DisplayDialog(options, script_path, job)
end
  
-- Did the user cancel
if dialog_code == 0 then
	return false
 end

-- Create contour from SBP file
local toolpath_contour = CreateContourFromSBPFile(options.filename)
if not toolpath_contour or toolpath_contour.IsEmpty then
  MessageBox("Error parsing toolpath")
  return false
end
    
-- Tell the user about the toolpath's dimensions
local bounds = toolpath_contour.BoundingBox3D
local format_spec = "Toolpath limits: %f x %f x %f"
local dimensions = string.format(format_spec, bounds.XLength, bounds.YLength, bounds.ZLength)
MessageBox(dimensions)

-- Update the start depth
-- If z zero was set to the table we adjust the start depth
if options.z_zero_set_to_table then
  local material_block = MaterialBlock()
  options.start_depth = material_block.Thickness
end

-- Require a group to create a toolpath
local toolpath_group = ContourGroup(true)
toolpath_group:AddTail(toolpath_contour)

-- Create an external toolpath
CreateToolpath(options, toolpath_group)

return true
end

